/*
 * vcard_funcs.h
 * vim: expandtab:ts=4:sts=4:sw=4
 *
 * Copyright (C) 2022 Marouane L. <techmetx11@disroot.org>
 *
 * This file is part of Profanity.
 *
 * Profanity is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Profanity is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Profanity.  If not, see <https://www.gnu.org/licenses/>.
 *
 * In addition, as a special exception, the copyright holders give permission to
 * link the code of portions of this program with the OpenSSL library under
 * certain conditions as described in each individual source file, and
 * distribute linked combinations including the two.
 *
 * You must obey the GNU General Public License in all respects for all of the
 * code used other than OpenSSL. If you modify file(s) with this exception, you
 * may extend this exception to your version of the file(s), but you are not
 * obligated to do so. If you do not wish to do so, delete this exception
 * statement from your version. If you delete this exception statement from all
 * source files in the program, then also delete it here.
 *
 */

#ifndef XMPP_VCARD_FUNCS_H
#define XMPP_VCARD_FUNCS_H

#include "ui/win_types.h"
#include "xmpp/vcard.h"

vCard* vcard_new();
void vcard_free(vCard* vcard);
void vcard_free_full(vCard* vcard);

gboolean vcard_parse(xmpp_stanza_t* vcard_xml, vCard* vcard);

void vcard_print(xmpp_ctx_t* ctx, ProfWin* window, char* jid);
void vcard_photo(xmpp_ctx_t* ctx, char* jid, char* filename, int index, gboolean open);

void vcard_user_refresh(void);
void vcard_user_save(void);
void vcard_user_set_fullname(char* fullname);
void vcard_user_set_name_family(char* family);
void vcard_user_set_name_given(char* given);
void vcard_user_set_name_middle(char* middle);
void vcard_user_set_name_prefix(char* prefix);
void vcard_user_set_name_suffix(char* suffix);

void vcard_user_add_element(vcard_element_t* element);
void vcard_user_remove_element(unsigned int index);
vcard_element_t* vcard_user_get_element_index(unsigned int index);
ProfWin* vcard_user_create_win();

void vcard_user_free(void);
#endif
